from contextlib import contextmanager
from typing import Iterator

from cleo.io.io import IO
from cleo.io.null_io import NullIO
from poetry.console.commands.command import Command
from poetry.console.commands.installer_command import InstallerCommand
from poetry.poetry import Poetry

from poetry_plugin_multiverse.cli.progress import progress
from poetry_plugin_multiverse.hooks.hook import Hook
from poetry_plugin_multiverse.repositories import create_installer, lock, locked_pool, workspace_pool
from poetry_plugin_multiverse.workspace import Workspace


@contextmanager
def poetry_target(poetry: Poetry, command: Command) -> Iterator[Command]:
    old_poetry = command.poetry
    try:
        command.set_poetry(poetry)
        yield command
    finally:
        command.set_poetry(old_poetry)


@contextmanager
def no_install(io: IO) -> Iterator[IO]:
    if io.input.has_option('lock'):
        old_value = io.input.option('lock')
        try:
            io.input.set_option('lock', True)
            yield io
        finally:
            io.input.set_option('lock', old_value)
    else:
        yield io


class PreLockHook(Hook):
    command_names = { 'add', 'lock', 'remove', 'update' }

    def run(self, workspace: Workspace, command: Command, io: IO):
        if not workspace.config.hooks_enabled('lock'):
            return
        if not isinstance(command, InstallerCommand):
            return

        root = workspace.root
        with progress(io, 'Preparing workspace...'):
            lock(
                root,
                locked_pool(*workspace.projects, packages=list(workspace.packages), strict=True)
            )

        root.set_pool(workspace_pool(*workspace.projects))
        command.set_installer(create_installer(
            root,
            root.pool
        ))
        with poetry_target(root, command) as cmd:
            with no_install(io) as lock_io:
                cmd.execute(NullIO(lock_io.input))

        command.poetry.set_pool(locked_pool(
            command.poetry,
            packages=root.locker.locked_repository().packages,
            strict=True
        ))
        command.set_installer(create_installer(
            command.poetry,
            command.poetry.pool,
            env=command.env,
            io=io
        ))


class PostLockHook(Hook):
    command_names = { 'add', 'lock', 'remove', 'update' }

    def run(self, workspace: Workspace, command: Command, io: IO):
        if workspace.config.hooks_enabled('lock'):
            command.call('workspace lock', '--no-update')



from contextlib import contextmanager
from typing import Iterator, List, Optional

from cleo.io.io import IO
from cleo.io.null_io import NullIO
from poetry.core.packages.dependency import Dependency
from poetry.core.packages.package import Package
from poetry.core.packages.utils.link import Link
from poetry.installation.installer import Installer
from poetry.poetry import Poetry
from poetry.repositories.repository import Repository
from poetry.repositories.repository_pool import RepositoryPool
from poetry.utils.env import Env

from poetry_plugin_multiverse.root import root_env
from poetry_plugin_multiverse.utils import Singleton


def repo_packages(upstream: Repository, packages: List[Package]) -> Iterator[Package]:
    for package in packages:
        if package.source_reference in { None, upstream.name }:
            yield package


class LockedRepository(Repository):
    def __init__(self, upstream: Repository, packages: List[Package], *, strict: bool) -> None:
        super().__init__(upstream.name, list(repo_packages(upstream, packages)))
        self.upstream = upstream
        self.strict = strict
    
    def find_packages(self, dependency: Dependency) -> List[Package]:
        if locked := super().find_packages(dependency):
            return locked
        if not self.strict:
            return self.upstream.find_packages(dependency)
        return []
    
    def has_package(self, package: Package) -> bool:
        if super().has_package(package):
            return True
        if not self.strict:
            return self.upstream.has_package(package)
        return False

    def search(self, *args, **kwargs) -> List[Package]:
        if locked := super().search(*args, **kwargs):
            return locked
        if not self.strict:
            return self.upstream.search(*args, **kwargs)
        return []
    
    def find_links_for_package(self, package: Package) -> List[Link]:
        return self.upstream.find_links_for_package(package)
    
    def package(self, *args, **kwargs) -> Package:
        return self.upstream.package(*args, **kwargs)


def non_empty_pools(*projects: Poetry) -> Iterator[RepositoryPool]:
    for poetry in projects:
        if len(poetry.pool.all_repositories) > 0:
            yield poetry.pool


def locked_pool(*projects: Poetry, packages: List[Package], strict: bool) -> RepositoryPool:
    pool = RepositoryPool()

    pools = list(non_empty_pools(*projects)) or [
        RepositoryPool.from_packages(packages, config=None)
    ]

    for project_pool in pools:
        for repo in project_pool.all_repositories:
            priority = project_pool.get_priority(repo.name)
            if not pool.has_repository(repo.name):
                locked = LockedRepository(repo, packages, strict=strict)
                pool.add_repository(locked, priority=priority)

    return pool


def workspace_pool(*projects: Poetry) -> RepositoryPool:
    return locked_pool(*projects, packages=[], strict=False)


def create_installer(
    project: Poetry, 
    pool: RepositoryPool, *,
    env: Optional[Env] = None,
    io: Optional[IO] = None
) -> Installer:
    return Installer(
        io or NullIO(),
        env or root_env(project),
        project.package,
        project.locker,
        pool,
        project.config,
        disable_cache=project.disable_cache
    )


def lock(project: Poetry, pool: RepositoryPool, *, env: Optional[Env] = None) -> int:
    return create_installer(project, pool, env=env).lock().run()


class PoolFactory(metaclass=Singleton):
    def __init__(self, pool: Optional[RepositoryPool] = None):
        self._pool = pool

    def get(self) -> Optional[RepositoryPool]:
        return self._pool
    
    @contextmanager
    def override(self, pool: RepositoryPool) -> Iterator[RepositoryPool]:
        old_pool = self._pool
        try:
            self._pool = pool
            yield pool
        finally:
            self._pool = old_pool



from abc import abstractmethod
from dataclasses import dataclass
from typing import Callable, ClassVar, Optional, Set
from cleo.events.event import Event
from cleo.events.console_event import ConsoleEvent
from cleo.io.io import IO
from poetry.console.commands.command import Command

from poetry_plugin_multiverse.workspace import Workspace


class Hook:
    command_names: ClassVar[Set[str]]

    @abstractmethod
    def run(self, workspace: Workspace, command: Command, io: IO):
        raise NotImplementedError()


@dataclass
class HookContext:
    workspace: Workspace
    command: Command
    io: IO

    def run(self, *hooks: Callable[[], Hook]):
        if not self.workspace.config.hooks_enabled():
            return
        for hook in hooks:
            hook_instance = hook()
            if self.command.name in hook_instance.command_names:
                hook_instance.run(self.workspace, self.command, self.io)

    @staticmethod
    def create(event: Event) -> Optional['HookContext']:
        if not isinstance(event, ConsoleEvent):
            return None
        if not isinstance(event.command, Command):
            return None
        if workspace := Workspace.create(event.command.poetry):
            return HookContext(workspace, event.command, event.io)
        return None



from contextlib import contextmanager
from typing import Iterator
from cleo.io.io import IO
from poetry.puzzle.provider import Indicator


@contextmanager
def progress(io: IO, message: str) -> Iterator[None]:
    if not io.output.is_decorated() or io.is_debug():
        io.write_line(message)
        yield
    else:
        indicator = Indicator(
            io, '{message}{context}<debug>({elapsed:2s})</debug>'
        )

        with indicator.auto(
            f'<info>{message}</info>',
            f'<info>{message}</info>'
        ):
            yield



from cleo.io.io import IO
from poetry.console.commands.command import Command

from poetry_plugin_multiverse.hooks.hook import Hook
from poetry_plugin_multiverse.publish import PublishedDependencies
from poetry_plugin_multiverse.workspace import Workspace


class PreBuildHook(Hook):
    command_names = { 'build' }

    def run(self, workspace: Workspace, command: Command, io: IO):
        if workspace.config.hooks_enabled('build'):
            PublishedDependencies(workspace.projects).patch_project(command.poetry)
